# Copyright 2018 The Cornac Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

# cython: language_level=3

from cython cimport floating, integral
from libcpp.vector cimport vector


cdef int get_key(int i_id, int j_id) nogil

cdef floating get_score(floating[:, :, :] G, int dim1, int dim2, int dim3,
                        floating[:, :] U, floating[:, :] I, floating[:, :] A,
                        int u_idx, int i_idx, int a_idx) nogil
